@@ -5,38 +5,12 @@ extern crate proc_macro;
 use proc_macro::TokenStream;
 
 use proc_macro2::{TokenStream as TokenStream2, TokenTree};
-use quote::{quote, ToTokens, TokenStreamExt};
+use quote::{quote, quote_spanned, ToTokens, TokenStreamExt};
 use syn::{
-    parse::{Parse, ParseStream},
-    parse_macro_input,
-    spanned::Spanned,
-    DeriveInput, Error, Generics, Ident, ItemFn, ItemType, LitStr, Visibility,
+    parse_macro_input, parse_quote, spanned::Spanned, Error, Fields, FnArg, Ident, ItemFn,
+    ItemStruct, LitStr, Pat, Visibility,
 };
 
-/// Parses a type definition, extracts its identifier and generic parameters
-struct TypeDefinition {
-    ident: Ident,
-    generics: Generics,
-}
-
-impl Parse for TypeDefinition {
-    fn parse(input: ParseStream) -> syn::Result<Self> {
-        if let Ok(d) = DeriveInput::parse(input) {
-            Ok(Self {
-                ident: d.ident,
-                generics: d.generics,
-            })
-        } else if let Ok(t) = ItemType::parse(input) {
-            Ok(Self {
-                ident: t.ident,
-                generics: t.generics,
-            })
-        } else {
-            Err(input.error("Input is not an alias, enum, struct or union definition"))
-        }
-    }
-}
-
 macro_rules! err {
     ($span:expr, $message:expr $(,)?) => {
         Error::new($span.span(), $message).to_compile_error()
@@ -46,28 +20,70 @@ macro_rules! err {
     };
 }
 
-/// `unsafe_guid` attribute macro, implements the `Identify` trait for any type
-/// (mostly works like a custom derive, but also supports type aliases)
+/// Attribute macro for marking structs as UEFI protocols.
+///
+/// The macro takes one argument, a GUID string.
+///
+/// The macro can only be applied to a struct, and the struct must have
+/// named fields (i.e. not a unit or tuple struct). It implements the
+/// [`Protocol`] trait and the `unsafe` [`Identify`] trait for the
+/// struct. It also adds a hidden field that causes the struct to be
+/// marked as [`!Send` and `!Sync`][send-and-sync].
+///
+/// # Safety
+///
+/// The caller must ensure that the correct GUID is attached to the
+/// type. An incorrect GUID could lead to invalid casts and other
+/// unsound behavior.
+///
+/// # Example
+///
+/// ```
+/// use uefi::{Identify, guid};
+/// use uefi::proto::unsafe_protocol;
+///
+/// #[unsafe_protocol("12345678-9abc-def0-1234-56789abcdef0")]
+/// struct ExampleProtocol {}
+///
+/// assert_eq!(ExampleProtocol::GUID, guid!("12345678-9abc-def0-1234-56789abcdef0"));
+/// ```
+///
+/// [`Identify`]: https://docs.rs/uefi/latest/uefi/trait.Identify.html
+/// [`Protocol`]: https://docs.rs/uefi/latest/uefi/proto/trait.Protocol.html
+/// [send-and-sync]: https://doc.rust-lang.org/nomicon/send-and-sync.html
 #[proc_macro_attribute]
-pub fn unsafe_guid(args: TokenStream, input: TokenStream) -> TokenStream {
-    // Parse the arguments and input using Syn
+pub fn unsafe_protocol(args: TokenStream, input: TokenStream) -> TokenStream {
+    // Parse `args` as a GUID string.
     let (time_low, time_mid, time_high_and_version, clock_seq_and_variant, node) =
         match parse_guid(parse_macro_input!(args as LitStr)) {
             Ok(data) => data,
             Err(tokens) => return tokens.into(),
         };
 
-    let mut result: TokenStream2 = input.clone().into();
+    let item_struct = parse_macro_input!(input as ItemStruct);
 
-    let type_definition = parse_macro_input!(input as TypeDefinition);
+    let ident = &item_struct.ident;
+    let struct_attrs = &item_struct.attrs;
+    let struct_vis = &item_struct.vis;
+    let struct_fields = if let Fields::Named(struct_fields) = &item_struct.fields {
+        &struct_fields.named
+    } else {
+        return err!(item_struct, "Protocol struct must used named fields").into();
+    };
+    let struct_generics = &item_struct.generics;
+    let (impl_generics, ty_generics, where_clause) = item_struct.generics.split_for_impl();
 
-    // At this point, we know everything we need to implement Identify
-    let ident = &type_definition.ident;
-    let (impl_generics, ty_generics, where_clause) = type_definition.generics.split_for_impl();
+    quote! {
+        #(#struct_attrs)*
+        #struct_vis struct #ident #struct_generics {
+            // Add a hidden field with `PhantomData` of a raw
+            // pointer. This has the implicit side effect of making the
+            // struct !Send and !Sync.
+            _no_send_or_sync: ::core::marker::PhantomData<*const u8>,
+            #struct_fields
+        }
 
-    result.append_all(quote! {
         unsafe impl #impl_generics ::uefi::Identify for #ident #ty_generics #where_clause {
-            #[doc(hidden)]
             const GUID: ::uefi::Guid = ::uefi::Guid::from_values(
                 #time_low,
                 #time_mid,
@@ -76,8 +92,39 @@ pub fn unsafe_guid(args: TokenStream, input: TokenStream) -> TokenStream {
                 #node,
             );
         }
-    });
-    result.into()
+
+        impl #impl_generics ::uefi::proto::Protocol for #ident #ty_generics #where_clause {}
+    }
+    .into()
+}
+
+/// Create a `Guid` at compile time.
+///
+/// # Example
+///
+/// ```
+/// use uefi::{guid, Guid};
+/// const EXAMPLE_GUID: Guid = guid!("12345678-9abc-def0-1234-56789abcdef0");
+/// ```
+#[proc_macro]
+pub fn guid(args: TokenStream) -> TokenStream {
+    let (time_low, time_mid, time_high_and_version, clock_seq_and_variant, node) =
+        match parse_guid(parse_macro_input!(args as LitStr)) {
+            Ok(data) => data,
+            Err(tokens) => return tokens.into(),
+        };
+
+    quote!({
+        const g: ::uefi::Guid = ::uefi::Guid::from_values(
+            #time_low,
+            #time_mid,
+            #time_high_and_version,
+            #clock_seq_and_variant,
+            #node,
+        );
+        g
+    })
+    .into()
 }
 
 fn parse_guid(guid_lit: LitStr) -> Result<(u32, u16, u16, u16, u64), TokenStream2> {
@@ -135,29 +182,61 @@ fn parse_guid(guid_lit: LitStr) -> Result<(u32, u16, u16, u16, u64), TokenStream
     ))
 }
 
-/// Custom derive for the `Protocol` trait
-#[proc_macro_derive(Protocol)]
-pub fn derive_protocol(item: TokenStream) -> TokenStream {
-    // Parse the input using Syn
-    let item = parse_macro_input!(item as DeriveInput);
-
-    // Then implement Protocol
-    let ident = item.ident.clone();
-    let (impl_generics, ty_generics, where_clause) = item.generics.split_for_impl();
-    let result = quote! {
-        // Mark this as a `Protocol` implementation
-        impl #impl_generics ::uefi::proto::Protocol for #ident #ty_generics #where_clause {}
-
-        // Most UEFI functions expect to be called on the bootstrap processor.
-        impl #impl_generics !Send for #ident #ty_generics #where_clause {}
-
-        // Most UEFI functions do not support multithreaded access.
-        impl #impl_generics !Sync for #ident #ty_generics #where_clause {}
-    };
-    result.into()
+/// Get the name of a function's argument at `arg_index`.
+fn get_function_arg_name(f: &ItemFn, arg_index: usize, errors: &mut TokenStream2) -> Option<Ident> {
+    if let Some(FnArg::Typed(arg)) = f.sig.inputs.iter().nth(arg_index) {
+        if let Pat::Ident(pat_ident) = &*arg.pat {
+            // The argument has a valid name such as `handle` or `_handle`.
+            Some(pat_ident.ident.clone())
+        } else {
+            // The argument is unnamed, i.e. `_`.
+            errors.append_all(err!(
+                arg.pat.span(),
+                "Entry method's arguments must be named"
+            ));
+            None
+        }
+    } else {
+        // Either there are too few arguments, or it's the wrong kind of
+        // argument (e.g. `self`).
+        //
+        // Don't append an error in this case. The error will be caught
+        // by the typecheck later on, which will give a better error
+        // message.
+        None
+    }
 }
 
-/// Custom attribute for a UEFI executable entrypoint
+/// Custom attribute for a UEFI executable entry point.
+///
+/// This attribute modifies a function to mark it as the entry point for
+/// a UEFI executable. The function must have two parameters, [`Handle`]
+/// and [`SystemTable<Boot>`], and return a [`Status`]. The function can
+/// optionally be `unsafe`.
+///
+/// Due to internal implementation details the parameters must both be
+/// named, so `arg` or `_arg` are allowed, but not `_`.
+///
+/// The [`BootServices::set_image_handle`] function will be called
+/// automatically with the image [`Handle`] argument.
+///
+/// # Examples
+///
+/// ```no_run
+/// #![no_main]
+///
+/// use uefi::prelude::*;
+///
+/// #[entry]
+/// fn main(image: Handle, st: SystemTable<Boot>) -> Status {
+///     Status::SUCCESS
+/// }
+/// ```
+///
+/// [`Handle`]: https://docs.rs/uefi/latest/uefi/data_types/struct.Handle.html
+/// [`SystemTable<Boot>`]: https://docs.rs/uefi/latest/uefi/table/struct.SystemTable.html
+/// [`Status`]: https://docs.rs/uefi/latest/uefi/struct.Status.html
+/// [`BootServices::set_image_handle`]: https://docs.rs/uefi/latest/uefi/table/boot/struct.BootServices.html#method.set_image_handle
 #[proc_macro_attribute]
 pub fn entry(args: TokenStream, input: TokenStream) -> TokenStream {
     // This code is inspired by the approach in this embedded Rust crate:
@@ -190,6 +269,9 @@ pub fn entry(args: TokenStream, input: TokenStream) -> TokenStream {
         ));
     }
 
+    let image_handle_ident = get_function_arg_name(&f, 0, &mut errors);
+    let system_table_ident = get_function_arg_name(&f, 1, &mut errors);
+
     // show most errors at once instead of one by one
     if !errors.is_empty() {
         return errors.into();
@@ -199,15 +281,55 @@ pub fn entry(args: TokenStream, input: TokenStream) -> TokenStream {
     let unsafety = f.sig.unsafety.take();
     // strip any visibility modifiers
     f.vis = Visibility::Inherited;
+    // Set the global image handle. If `image_handle_ident` is `None`
+    // then the typecheck is going to fail anyway.
+    if let Some(image_handle_ident) = image_handle_ident {
+        f.block.stmts.insert(
+            0,
+            parse_quote! {
+                unsafe {
+                    #system_table_ident.boot_services().set_image_handle(#image_handle_ident);
+                }
+            },
+        );
+    }
 
-    let ident = &f.sig.ident;
+    let fn_ident = &f.sig.ident;
+    // Get an iterator of the function inputs types. This is needed instead of
+    // directly using `sig.inputs` because patterns you can use in fn items like
+    // `mut <arg>` aren't valid in fn pointers.
+    let fn_inputs = f.sig.inputs.iter().map(|arg| match arg {
+        FnArg::Receiver(arg) => quote!(#arg),
+        FnArg::Typed(arg) => {
+            let ty = &arg.ty;
+            quote!(#ty)
+        }
+    });
+    let fn_output = &f.sig.output;
+    let signature_span = f.sig.span();
+
+    let fn_type_check = quote_spanned! {signature_span=>
+        // Cast from the function type to a function pointer with the same
+        // signature first, then try to assign that to an unnamed constant with
+        // the desired function pointer type.
+        //
+        // The cast is used to avoid an "expected fn pointer, found fn item"
+        // error if the signature is wrong, since that's not what we are
+        // interested in here. Instead we want to tell the user what
+        // specifically in the function signature is incorrect.
+        const _:
+            // The expected fn pointer type.
+            #unsafety extern "efiapi" fn(::uefi::Handle, ::uefi::table::SystemTable<::uefi::table::Boot>) -> ::uefi::Status =
+            // Cast from a fn item to a function pointer.
+            #fn_ident as #unsafety extern "efiapi" fn(#(#fn_inputs),*) #fn_output;
+    };
 
     let result = quote! {
+        #fn_type_check
+
         #[export_name = "efi_main"]
         #unsafety extern "efiapi" #f
 
-        // typecheck the function pointer
-        const _: #unsafety extern "efiapi" fn(::uefi::Handle, ::uefi::table::SystemTable<::uefi::table::Boot>) -> ::uefi::Status = #ident;
     };
     result.into()
 }
